/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aacdec_filter.c paraslash's aac (m4a) decoder. */

#include "para.h"

#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "filter.h"
#include "error.h"
#include "string.h"
#include "aac.h"

/** the output buffer size */
#define AAC_OUTBUF_SIZE (32 * 1024)

/** give up decoding after that many errors */
#define MAX_ERRORS 20

/**
 * data specific to the aacdec filter
 *
 * \sa filter, filter_node
 */
struct private_aacdec_data {
	/** the return value of aac_open */
	NeAACDecHandle handle;
	/** info about the currently decoded frame */
	NeAACDecFrameInfo frame_info;
	/** whether this instance of the aac decoder is already initialized */
	int initialized;
	/**
	 * return value of aac_find_esds(). Used to call the right aacdec
	 * init function
	 */
	unsigned long decoder_length;
	/** number of times the decoder returned an error */
	unsigned error_count;
	/** number of bytes already consumed from the imput stream */
	size_t consumed_total;
	/** return value of aac_find_entry_point */
	size_t entry;
};

static ssize_t aacdec(char *input_buffer, size_t len, struct filter_node *fn)
{
	struct private_aacdec_data *padd = fn->private_data;
	struct filter_chain *fc = fn->fc;
	int i, ret;
	unsigned char *p, *outbuffer;
	unsigned char *inbuf = (unsigned char*)input_buffer;
	size_t skip, consumed = 0;

	if (fn->loaded > fn->bufsize * 3 / 5)
		return 0;
	ret = *fc->input_error;
	if (ret < 0)
		return ret;
	if (len < 2048)
		return 0;

	if (!padd->initialized) {
		unsigned long rate = 0;
		unsigned char channels = 0;
		ret = aac_find_esds(inbuf, len, &skip, &padd->decoder_length);
		if (ret < 0) {
			PARA_INFO_LOG("%s\n", para_strerror(-ret));
			ret = NeAACDecInit(padd->handle, inbuf,
				len, &rate, &channels);
			PARA_INFO_LOG("decoder init: %d\n", ret);
			if (ret < 0) {
				ret = -E_AACDEC_INIT;
				goto out;
			}
			consumed = ret;
		} else {
			PARA_INFO_LOG("decoder len: %lu\n",
				padd->decoder_length);
			consumed += skip;
			p = inbuf + consumed;
			ret = -E_AACDEC_INIT;
			if (NeAACDecInit2(padd->handle, p,
					padd->decoder_length, &rate,
					&channels) < 0)
				goto out;
		}
		fc->samplerate = rate;
		fc->channels = channels;
		PARA_INFO_LOG("rate: %u, channels: %d\n",
			fc->samplerate, fc->channels);
		padd->initialized = 1;
	}
	if (padd->decoder_length > 0) {
		consumed = 0;
		if (!padd->entry) {
			ret = aac_find_entry_point(inbuf + consumed,
				len - consumed, &skip);
			if (ret < 0) {
				ret = len;
				goto out;
			}
			consumed += skip;
			padd->entry = ret;
			PARA_INFO_LOG("entry: %zu\n", padd->entry);
		}
		ret = len;
		if (padd->consumed_total + len < padd->entry)
			goto out;
		if (padd->consumed_total < padd->entry)
			consumed = padd->entry - padd->consumed_total;
	}
	for (; consumed < len; consumed++)
		if ((inbuf[consumed] & 0xfe) == 0x20)
			break;
	if (consumed >= len)
		goto success;
	p = inbuf + consumed;
	outbuffer = NeAACDecDecode(padd->handle, &padd->frame_info, p,
		len - consumed);
	if (padd->frame_info.error) {
		ret = -E_AAC_DECODE;
		if (padd->error_count++ > MAX_ERRORS)
			goto out;
		PARA_ERROR_LOG("frame_error: %d, consumed: %zu + %zd + %lu\n",
			padd->frame_info.error, padd->consumed_total,
			consumed, padd->frame_info.bytesconsumed);
		PARA_ERROR_LOG("%s\n", NeAACDecGetErrorMessage(
			padd->frame_info.error));
		consumed++; /* catch 21 */
		goto success;
	}
	padd->error_count = 0;
	consumed += padd->frame_info.bytesconsumed;
	ret = consumed;
	if (!padd->frame_info.samples)
		goto out;
	ret = -E_AAC_OVERRUN;
	if (padd->frame_info.samples * 2 + fn->loaded > fn->bufsize)
		goto out;
	for (i = 0; i < padd->frame_info.samples; i++) {
		short *s = (short *)outbuffer;
		write_int16_host_endian(fn->buf + fn->loaded, s[i]);
		fn->loaded += 2;
	}
success:
	ret = consumed;
out:
	if (ret > 0)
		padd->consumed_total += ret;
	return ret;
}

static void aacdec_open(struct filter_node *fn)
{
	struct private_aacdec_data *padd = para_calloc(sizeof(*padd));

	fn->private_data = padd;
	fn->bufsize = AAC_OUTBUF_SIZE;
	fn->buf = para_calloc(fn->bufsize);
	padd->handle = aac_open();
}

static void aacdec_close(struct filter_node *fn)
{
	struct private_aacdec_data *padd = fn->private_data;

	NeAACDecClose(padd->handle);
	free(fn->buf);
	fn->buf = NULL;
	free(padd);
	fn->private_data = NULL;
}

/**
 * the init function of the aacdec filter
 *
 * \param f pointer to the filter struct to initialize
 *
 * \sa filter::init
 */
void aacdec_filter_init(struct filter *f)
{
	f->open = aacdec_open;
	f->convert = aacdec;
	f->close = aacdec_close;
}
