/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file alsa_write.c paraslash's alsa output plugin */

/*
 * Based in parts on aplay.c from the alsa-utils-1.0.8 package,
 * Copyright (c) by Jaroslav Kysela <perex@suse.cz>, which is
 * based on the vplay program by Michael Beck.
 */

#include <sys/types.h>
#include <dirent.h>
#include <alsa/asoundlib.h>

#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "write.h"
#include "alsa_write.cmdline.h"
#include "error.h"

/** always use 16 bit little endian */
#define FORMAT SND_PCM_FORMAT_S16_LE

/** Data specific to the alsa writer. */
struct private_alsa_write_data {
	/** The alsa handle */
	snd_pcm_t *handle;
	/** Determined and set by alsa_open(). */
	int bytes_per_frame;
	/** Don't write anything until this time. */
	struct timeval next_chunk;
	/** The approximate maximum buffer duration in us. */
	unsigned buffer_time;
	/**
	 * The samplerate given by command line option or the decoder
	 * of the writer node group.
	 */
	unsigned samplerate;
	/**
	 * The number of channels, given by command line option or the
	 * decoder of the writer node group.
	 */
	unsigned channels;
};

/* Install PCM software and hardware configuration. */
static int alsa_init(struct private_alsa_write_data *pad,
		struct alsa_write_args_info *conf)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t buffer_size, start_threshold, stop_threshold;
	snd_pcm_uframes_t period_size;
	int err;

	err = snd_pcm_open(&pad->handle, conf->device_arg,
		SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);
	if (err < 0)
		return -E_PCM_OPEN;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(pad->handle, hwparams) < 0)
		return -E_BROKEN_CONF;
	if (snd_pcm_hw_params_set_access(pad->handle, hwparams,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		return -E_ACCESS_TYPE;
	if (snd_pcm_hw_params_set_format(pad->handle, hwparams, FORMAT) < 0)
		return -E_SAMPLE_FORMAT;
	if (snd_pcm_hw_params_set_channels(pad->handle, hwparams,
			pad->channels) < 0)
		return -E_CHANNEL_COUNT;
	if (snd_pcm_hw_params_set_rate_near(pad->handle, hwparams,
			&pad->samplerate, 0) < 0)
		return -E_SET_RATE;
	err = snd_pcm_hw_params_get_buffer_time_max(hwparams, &pad->buffer_time, 0);
	if (err < 0 || !pad->buffer_time)
		return -E_GET_BUFFER_TIME;
	PARA_INFO_LOG("buffer time: %d\n", pad->buffer_time);
	if (snd_pcm_hw_params_set_buffer_time_near(pad->handle, hwparams,
			&pad->buffer_time, 0) < 0)
		return -E_SET_BUFFER_TIME;
	if (snd_pcm_hw_params(pad->handle, hwparams) < 0)
		return -E_HW_PARAMS;
	snd_pcm_hw_params_get_period_size(hwparams, &period_size, 0);
	snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size);
	PARA_INFO_LOG("buffer size: %lu, period_size: %lu\n", buffer_size,
		period_size);
	if (period_size == buffer_size)
		return -E_BAD_PERIOD;
	snd_pcm_sw_params_current(pad->handle, swparams);
	snd_pcm_sw_params_set_avail_min(pad->handle, swparams, period_size);
	if (buffer_size < 1)
		start_threshold = 1;
	else
		start_threshold = PARA_MIN(buffer_size,
			(snd_pcm_uframes_t)pad->samplerate);
	if (snd_pcm_sw_params_set_start_threshold(pad->handle, swparams,
			start_threshold) < 0)
		return -E_START_THRESHOLD;
	stop_threshold = buffer_size;
	if (snd_pcm_sw_params_set_stop_threshold(pad->handle, swparams,
			stop_threshold) < 0)
		return -E_STOP_THRESHOLD;
	if (snd_pcm_sw_params(pad->handle, swparams) < 0)
		PARA_WARNING_LOG("unable to install sw params\n");
	pad->bytes_per_frame = snd_pcm_format_physical_width(FORMAT)
		* pad->channels / 8;
	if (pad->bytes_per_frame <= 0)
		return -E_PHYSICAL_WIDTH;
	PARA_INFO_LOG("bytes per frame: %d\n", pad->bytes_per_frame);
	if (snd_pcm_nonblock(pad->handle, 1))
		PARA_ERROR_LOG("failed to set nonblock mode\n");
	return 1;
}

/* Open an instance of the alsa writer. */
static int alsa_open(struct writer_node *wn)
{
	struct alsa_write_args_info *conf = wn->conf;
	struct writer_node_group *wng = wn->wng;
	struct private_alsa_write_data *pad = para_calloc(sizeof(*pad));

	wn->private_data = pad;
	if (!conf->samplerate_given && wng->samplerate)
		pad->samplerate = *wng->samplerate;
	else
		pad->samplerate = conf->samplerate_arg;
	if (!conf->channels_given && wng->channels)
		pad->channels = *wng->channels;
	else
		pad->channels = conf->channels_arg;
	PARA_INFO_LOG("%d channel(s), %dHz\n", pad->channels, pad->samplerate);
	tv_add(now, &(struct timeval){0, 100 * 1000}, &pad->next_chunk);
	return 1;
}

static int alsa_write_pre_select(struct sched *s, struct writer_node *wn)
{
	struct private_alsa_write_data *pad = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	struct timeval diff;

	if (!*wng->loaded)
		return 1;
	if (tv_diff(now, &pad->next_chunk, &diff) < 0) {
		if (tv_diff(&s->timeout, &diff, NULL) > 0)
			s->timeout = diff;
	} else {
		s->timeout.tv_sec = 0;
		s->timeout.tv_usec = 1;
	}
	return 1;
}

static int alsa_write_post_select(__a_unused struct sched *s,
		struct writer_node *wn)
{
	struct private_alsa_write_data *pad = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	size_t frames, bytes = *wng->loaded - wn->written;
	unsigned char *data = (unsigned char*)wng->buf + wn->written;
	struct timeval tv;
	snd_pcm_sframes_t ret;

	if (*wng->input_error < 0 && (!pad->handle || bytes < pad->bytes_per_frame)) {
		wn->written = *wng->loaded;
		return *wng->input_error;
	}
	if (!bytes) /* no data available */
		goto out;
	if (tv_diff(now, &pad->next_chunk, NULL) < 0)
		goto out;
	if (!pad->handle) {
		int err = alsa_init(pad, wn->conf);
		if (err < 0)
			return err;
	}
	frames = bytes / pad->bytes_per_frame;
	if (!frames) /* less than a single frame available */
		goto out;
	ret = snd_pcm_writei(pad->handle, data, frames);
	if (ret == -EPIPE) {
		PARA_WARNING_LOG("EPIPE\n");
		snd_pcm_prepare(pad->handle);
		return 1;
	}
	if (ret < 0 && ret != -EAGAIN) {
		PARA_WARNING_LOG("alsa error (%zu frames, ret = %d\n",
			frames, (int)ret);
		return -E_ALSA_WRITE;
	}
	if (ret == -EAGAIN)
		PARA_DEBUG_LOG("EAGAIN\n");
	else
		wn->written += ret * pad->bytes_per_frame;
	if (ret == frames) /* we wrote everything, try again immediately */
		pad->next_chunk	= *now;
	else { /* wait until 50% buffer space is available */
		ms2tv(pad->buffer_time / 2000, &tv);
		tv_add(now, &tv, &pad->next_chunk);
	}
out:
	return 1;
}

static void alsa_close(struct writer_node *wn)
{
	struct private_alsa_write_data *pad = wn->private_data;
	PARA_INFO_LOG("closing writer node %p\n", wn);

	if (pad->handle) {
		snd_pcm_drain(pad->handle);
		snd_pcm_close(pad->handle);
		snd_config_update_free_global();
	}
	free(pad);
}

__malloc static void *alsa_parse_config(const char *options)
{
	int ret;
	struct alsa_write_args_info *conf
		= para_calloc(sizeof(struct alsa_write_args_info));

	PARA_INFO_LOG("options: %s, %zd\n", options, strcspn(options, " \t"));
	ret = alsa_cmdline_parser_string(options, conf, "alsa_write");
	if (ret)
		goto err_out;
	PARA_INFO_LOG("help given: %d\n", conf->help_given);
	return conf;
err_out:
	free(conf);
	return NULL;
}

/**
 * the init function of the alsa writer
 *
 * \param w pointer to the writer to initialize
 *
 * \sa struct writer
 */
void alsa_write_init(struct writer *w)
{
	struct alsa_write_args_info dummy;

	alsa_cmdline_parser_init(&dummy);
	w->open = alsa_open;
	w->close = alsa_close;
	w->pre_select = alsa_write_pre_select;
	w->post_select = alsa_write_post_select;
	w->parse_config = alsa_parse_config;
	w->shutdown = NULL; /* nothing to do */
	w->help = (struct ggo_help) {
		.short_help = alsa_write_args_info_help,
		.detailed_help = alsa_write_args_info_detailed_help
	};
}
