/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter.h Filter-related structures and exported symbols from filter_common.c. */

/** The list of supported filters. */
enum filter_enum {FILTER_ENUM};

/**
 * Describes one running instance of a filter.
*/
struct filter_node {
	/** The number in the array of available filters. */
	unsigned filter_num;
	/** The filter chain this filter node belongs to. */
	struct filter_chain *fc;
	/**
	 * Each filter may store any filter-specific information about the particular
	 * instance of the filter here.
	 */
	void *private_data;
	/** The output buffer. */
	char *buf;
	/** The size of the output buffer. */
	size_t bufsize;
	/** The number of bytes currently loaded in \a buf. */
	size_t loaded;
	/** The list of registered callbacks. */
	struct list_head callbacks;
	/** A pointer to the configuration of this instance. */
	void *conf;
};

/** Describes one running instance of a chain of filters */
struct filter_chain {
	/** The length of the filter chain. */
	unsigned int num_filters;
	/**
	 * The number of channels of the current stream.
	 *
	 * Set by the decoding filter.
	 */
	unsigned int channels;
	/**
	 * Current sample rate in Hz.
	 *
	 * Set by the decoding filter.
	 */
	unsigned int samplerate;
	/** The list containing all filter nodes in this filter chain. */
	struct filter_node *filter_nodes;
	/**
	 * The input buffer of the filter chain.
	 *
	 * This is set to point to the output buffer of the receiving application (the
	 * buffer used to read from stdin for para_filter; the output buffer of the
	 * current receiver for para_audiod).
	 */
	char *inbuf;
	/**
	 * The output buffer of the filter chain.
	 *
	 * Points to the output buffer of the last filter in the filter chain.
	 */
	char *outbuf;
	/** Contains the number of bytes loaded in the input buffer. */
	size_t *in_loaded;
	/** Contains the number of bytes loaded in the output buffer. */
	size_t *out_loaded;
	/** Pointer to the error variable of the receiving application. */
	int *input_error;
	/** Pointer to the error variable of the writing application. */
	int *output_error;
	/** The task associated with the filter chain. */
	struct task task;
};

#define FOR_EACH_FILTER_NODE(fn, fc, i) for (i = 0; i < (fc)->num_filters \
	&& (fn = (fc)->filter_nodes + i); i++)


/**
 * Used to manage grab clients.
 *
 * An application using paraslash's filter subsystem may register any number of
 * callbacks for each filter_node. It is possible to attach a filter callback
 * while the filter is running. This is used for stream grabbing in
 * para_audiod: Whenever a client sends the 'grab' command, para_audiod adds a
 * filter callback to the list of callbacks for the filter node specified in
 * the grab command.
 */
struct filter_callback {
	/** All callbacks are organized in a doubly linked list. */
	struct list_head node;
	/**
	 * Private data.
	 *
	 * May be initialized by the application before registering the callback. This
	 * pointer is not used by the filter subsystem. It is provided for use within
	 * the input/output/close callback functions.
	 */
	void *data;
	/**
	 * The input callback.
	 *
	 * In not \p NULL, the filter subsystem calls this function whenever the filter
	 * consumed some or all of its input buffer. A pointer to the buffer of consumed
	 * data, its length and a pointer to the own \a filter_callback structure are passed
	 * to \a input_cb. The input callback is expected to return a negative value on errors.
	 */
	int (*input_cb)(char *buf, size_t len, struct filter_callback *fc);
	/**
	 * The output callback.
	 *
	 * If not NULL, this is called whenever the filter produces output. A pointer
	 * to the output data, its length and a pointer to the own \a filter_callback
	 * structure are passed to \a output_cb. Like the input callback, the output
	 * callback is expected to return a negative value on errors.
	 */
	int (*output_cb)(char *buf, size_t len, struct filter_callback *fc);
	/**
	 * The callback close function.
	 *
	 * This gets called whenever the input/output callback returned an error, or if
	 * the filter chain is going to be destroyed, e.g. because the end of the
	 * stream was encountered. It is assumed to succeed.
	 */
	void (*close)(struct filter_callback *fc);
};


/**
 * The structure associated with a paraslash filter.
 *
 * Paraslash filters are "modules" which are used to transform an audio stream.
 * struct filter contains pointers to functions that must be supplied by the
 * filter code in order to be used by the driving application (currently
 * para_audiod and para_filter).
 *
 * Note: As several instances of the same filter may be running at the same
 * time, all these filter functions must be reentrant; no static non-constant
 * variables may be used.
 * \sa mp3dec_filter.c, oggdec_filter.c, wav_filter.c, compress_filter.c, filter_node
 */
struct filter {
	/** The name of the filter. */
	const char *name;
	/**
	 * Pointer to the filter init routine.
	 *
	 * This function is only called once at startup. It must initialize the
	 * other non-optional function pointers of this structure.
	 */
	void (*init)(struct filter *f);
	/**
	 * Open one instance of this filter.
	 *
	 * This should allocate the output buffer of the given filter node and do any
	 * other filter-specific preparations like initializing the private_data member
	 * of \a fn suitably. The open function is assumed to succeed.
	 */
	void (*open)(struct filter_node *fn);
	/**
	 * Convert (filter) the given data.
	 *
	 * Pointer to the converting function of the filter. It should convert the
	 * given input buffer \a inbuf which is of length \a len to the previously
	 * reserved output buffer of \a fn. On success, it must return the number of
	 * bytes it consumed from \a inbuf. On errors, a negative number indicating the
	 * kind of the error must be returned.
	 *
	 * A zero return value just means that nothing was converted (probably because
	 * the input buffer was too small). This is not interpreted as an error.
	 */
	ssize_t (*convert)(char *inbuf, size_t len, struct filter_node *fn);
	/**
	 * Close one instance of this filter.
	 *
	 * Free all resources of associated with \a fn that were previously allocated
	 * by the open() function.
	 */
	void (*close)(struct filter_node *fn);
	/**
	 * A pointer to the filter's command line parser.
	 *
	 * If this optional function pointer is not NULL, any filter options
	 * are passed from the main program to this command line parser once at
	 * application startup. The command line parser should check its
	 * command line options given by \a argc and \a argv and abort on
	 * errors. Success must be indicated by a non-negative return value. In
	 * this case the function should return a pointer to the
	 * filter-specific configuration data determined by \a argc and \a
	 * argv. On failure, a negative paraslash error code must be returned.
	 */
	int (*parse_config)(int argc, char **argv, void **config);

	/** The help texts for this filter. */
	struct ggo_help help;
};

void close_filters(struct filter_chain *fc);
void filter_init(void);
int check_filter_arg(char *filter_arg, void **conf);
void filter_pre_select(__a_unused struct sched *s, struct task *t);
void print_filter_helps(int detailed);

static inline void write_int16_host_endian(char *buf, int val)
{
#ifdef WORDS_BIGENDIAN
	*buf = val >> 8;
	*(buf + 1) = val & 0xff;
#else
	*buf = val & 0xff;
	*(buf + 1) = val >> 8;
#endif
}

DECLARE_FILTER_INITS

#define FOR_EACH_SUPPORTED_FILTER(j)  for (j = 0; j < NUM_SUPPORTED_FILTERS; j++)

/** The filter array, one structure for each supported filter. */
extern struct filter filters[NUM_SUPPORTED_FILTERS];
